/*==========================================================================
 *
 *  Copyright (C) 1995-1997 Microsoft Corporation. All Rights Reserved.
 *  Copyright (C) 1994-1995 ATI Technologies Inc. All Rights Reserved.
 *
 *  File:	ddraw.c
 *  Content:	Misc. Direct Draw access routines
 *
 ***************************************************************************/
#include "winstuff.h"
#include "d_ddraw.h"
#include "f:\directx\sdk\inc\ddraw.h"
#include "doomdef.h"

PALETTEENTRY    ape[256];
IDirectDrawPalette *ppal=0;
LPDIRECTDRAW	lpDD;
LPDIRECTDRAW2   lpDD2;
LPDIRECTDRAWSURFACE  lpFrontBuffer;
LPDIRECTDRAWSURFACE  lpBackBuffer;
LPDIRECTDRAWSURFACE      lpStretchBuffer;
LPDIRECTDRAWCLIPPER      lpClipper;
BOOL                     bModeX=0;         // we are in a modex mode
int                      nBufferCount=0;   // buffer count
BOOL                     bFullscreen=1;    // run in fullscreen mode
int                      dispwidth=SCREENWIDTH;
BOOL DDCreateFlippingSurface( void );
BOOL DDClear( void );
void SetHiPal(int bpp);
int useHiPal = 0;

typedef struct {
    GUID *guid;
    GUID DriverGUID;
    CHAR DriverDesc[128];
    CHAR DriverName[128];
} Devices;

Devices aDDDevs[15];
int     MaxDevIndex = 0;
int     DevIndex = 0;

BOOL    bUseEmulation;
BOOL    bUseSysMem;
int     nBufferCount;
int     CmdLineBufferCount;
BOOL	bTransDest;
BOOL    bColorFill;

/*
 * DDInit
 */
/*
 * DDEnable
 */
static void
ReportError(char *Message, HRESULT hRes)
{
    HDC  hdc;
    char szBuffer[256];
    char szMessage[128];
    char *szError;
    int  nStrID;



    switch (hRes)
    {
        case DDERR_EXCEPTION:        szError = "ERR_EXCEPTION";        break;
        case DDERR_GENERIC:          szError = "ERR_GENERIC";          break;
        case DDERR_OUTOFMEMORY:      szError = "ERR_OUTOFMEMORY";      break;
        case DDERR_OUTOFVIDEOMEMORY: szError = "ERR_OUTOFVIDEOMEMORY"; break;
        case DDERR_SURFACEBUSY:      szError = "ERR_SURFACEBUSY";      break;

        case DDERR_SURFACELOST:      szError = "ERR_SURFACELOST";      break;
        case DDERR_WRONGMODE:        szError = "ERR_WRONGMODE";        break;
        default:                     szError = "ERR_INTERNALERROR";    break;
    }

    // Convert the error code into a string.


    wsprintf(szBuffer, "%s\n%s (Error #%d)", Message, szError,(hRes));
    I_Error(szBuffer);
}


BOOL DD_Init( void )
{
    LPDIRECTDRAW	lpdd;
    LPDIRECTDRAW2   lpdd2;
    DDCAPS          ddcaps;
    HRESULT		ddrval;
    BOOL		use_dest;
	OSVERSIONINFO osv;
    RECT rcWork;
    RECT rc;
    HDC hdc;
    DWORD dwStyle;
    int bpp=8;
    //Print("DirectDraw");

    if ((lpDD == NULL) || (lpDD2 == NULL))
    {
        if( bUseEmulation)
            DevIndex = MaxDevIndex-1;
        ddrval = DirectDrawCreate(aDDDevs[DevIndex].guid,&lpdd,NULL);
        if (lpDD == NULL){
            ddrval = IDirectDraw_SetCooperativeLevel( lpdd, mainhwnd,
                                DDSCL_ALLOWMODEX |
                                DDSCL_EXCLUSIVE |
    						    DDSCL_FULLSCREEN );


            if( ddrval != DD_OK ){
                ReportError("SetCooperativeLevel failed err", ddrval);
                goto error;
            }
        }
        if(ddrval == DD_OK)
        {
            ddrval = IDirectDraw_QueryInterface(lpdd, &IID_IDirectDraw2, &lpdd2);
        }
    }
    else
    {
        lpdd = lpDD;
        lpdd2 = lpDD2;
        ddrval = DD_OK;
    }

    if( ddrval != DD_OK )
    {
        ReportError("DirectDrawCreate failed err", ddrval);
        goto error;
    }

    /*
     * grab exclusive mode if we are going to run as fullscreen
     * otherwise grab normal mode.
     */

    if (bFullscreen)
    {
        //I_Error("SetDisplayMode %d %d %d",GameMode.cx,GameMode.cy, GameBPP);
        //if((GameMode.cx == 320) && (GameMode.cy == 200) && (GameBPP == 8))
        //{
        ddrval = IDirectDraw2_SetDisplayMode(lpdd2, SCREENWIDTH, SCREENHEIGHT, 8,0,DDSDM_STANDARDVGAMODE);
        //} else {
        //ddrval = IDirectDraw_SetDisplayMode( lpdd, SCREENWIDTH, SCREENHEIGHT, 8);
        //}
        if (ddrval != DD_OK && (SCREENWIDTH != 640 || SCREENHEIGHT != 480))
        {
            Print( "cant set mode trying 640x480" );
            ddrval = IDirectDraw_SetDisplayMode(lpdd, 640, 480, 8);
        }
/*        if(ddrval !=DD_OK){

            //
            //  when in rome (I mean when in windows) we should use the
            //  current mode
            //
            Print("Windowed mode because we can't get one of the others.");
            bFullscreen=0;
            ddrval = IDirectDraw_SetCooperativeLevel( lpdd, mainhwnd, DDSCL_NORMAL );


            hdc = GetDC(NULL);
            bpp = GetDeviceCaps(hdc, PLANES) * GetDeviceCaps(hdc, BITSPIXEL);
            ReleaseDC(NULL, hdc);
            //make the high palette
            useHiPal= bpp;
            SetRect(&rc, 0, 0, SCREENWIDTH, SCREENHEIGHT);

            AdjustWindowRectEx(&rc,
                NULL,
                GetMenu(mainhwnd) != NULL,
                NULL);

            SetWindowPos(mainhwnd, NULL, 0, 0, rc.right-rc.left, rc.bottom-rc.top,
                SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);

            SetWindowPos(mainhwnd, HWND_NOTOPMOST, 0, 0, 0, 0,
                SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE);

            //
            //  make sure our window does not hang outside of the work area
            //  this will make people who have the tray on the top or left
            //  happy.
            //
            SystemParametersInfo(SPI_GETWORKAREA, 0, &rcWork, 0);
            GetWindowRect(mainhwnd, &rc);
            if (rc.left < rcWork.left) rc.left = rcWork.left;
            if (rc.top  < rcWork.top)  rc.top  = rcWork.top;
            SetWindowPos(mainhwnd, NULL, rc.left, rc.top, 0, 0,
                SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
            ddrval=DD_OK;
        }   //Failed attempt to make a windowed mode.
            //maybe next time.
        */

        if( ddrval != DD_OK )
        {
            ReportError("SetMode failed err ", ddrval);
            goto error;
        }
    }

    /*
     * check capabilites
     */
    ddcaps.dwSize = sizeof( ddcaps );
    ddrval = IDirectDraw_GetCaps( lpdd, &ddcaps, NULL );

    if( ddrval != DD_OK )
    {
        ReportError("GetCaps failed err", ddrval);
        goto error;
    }

    /*
     * default to double buffered on 1mb, triple buffered
     * on > 1mb
     */
    if (nBufferCount == 0)
    {
        if( ddcaps.dwVidMemTotal <= 1024L*1024L*(8/8) ||
            SCREENWIDTH > 640 || bFullscreen)
        {
            //I_Error("double buffering (not enough memory)");
            nBufferCount = 2;
        }
        else
        {
            //I_Error("triple buffering");
            nBufferCount = 3;
        }
    }

    lpDD = lpdd;
    lpDD2 = lpdd2;
    DDCreateFlippingSurface();
    return TRUE;

error:
    return FALSE;

} /* DDEnable */

/*
 * DDDisable
 */
BOOL DDDisable( BOOL fFinal )
{
    Print("D_DRAW: Closing out");
    if( lpClipper )
    {
        IDirectDrawClipper_Release(lpClipper);
        lpClipper = NULL;
    }

    if( lpBackBuffer )
    {
        IDirectDrawSurface_Release(lpBackBuffer);
        lpBackBuffer = NULL;
    }

    if( lpFrontBuffer )
    {
        IDirectDrawSurface_Release(lpFrontBuffer);
        lpFrontBuffer = NULL;
    }

    if( lpStretchBuffer )
    {
        IDirectDrawSurface_Release(lpStretchBuffer);
        lpStretchBuffer = NULL;
    }

    //
    // fFinal is TRUE when the app is exiting, FALSE if we are
    // just seting a new game size..
    //
    if ( fFinal )
    {
        if( lpDD != NULL )
        {
            //Print("Release 1");
            if(lpDD2)
                IDirectDraw_Release( lpDD2 );
            IDirectDraw_Release( lpDD );
            //Print("Release 2");
            //Print("Release done");
            lpDD = NULL;
        }
    }

    return TRUE;
}

void DD_Stop(){
    DDDisable(1);
}

/*
 * updateFrame
 *
 * Copy the Doom frame buffer
 */
void updateFrame(char *Frame)
{
    DDBLTFX     ddbltfx;
    int		i;
    HRESULT     ddrval;
    DDSURFACEDESC Surf;
    char *ddbb,*doomscr,mys[128];
    //Print("Start Frame Swap");

    UpdateWindow(mainhwnd);

    Surf.dwSize = sizeof(DDSURFACEDESC);
    Surf.dwFlags = 0;
    if (bFullscreen<100)
    {
        {
            ddrval = IDirectDrawSurface_Lock(lpBackBuffer,NULL,&Surf,DDLOCK_SURFACEMEMORYPTR,NULL);
            while(ddrval == DDERR_WASSTILLDRAWING||ddrval==DDERR_SURFACEBUSY)//so it doesn't die if it is too quick.
                ddrval = IDirectDrawSurface_Lock(lpBackBuffer,NULL,&Surf,DDLOCK_SURFACEMEMORYPTR,NULL);
            if( ddrval != DD_OK )
            {
                ReportError("Lock failed ddrval = ", ddrval);
                return ;
            }
            if(SCREENWIDTH==dispwidth){
                memcpy(Surf.lpSurface,Frame,SCREENWIDTH*SCREENHEIGHT);
            }
            else{
                doomscr = Frame;
                ddbb = Surf.lpSurface;
                for(i=0;i<SCREENHEIGHT;i++){
                    memcpy(ddbb,doomscr,SCREENWIDTH);
                    doomscr+=SCREENWIDTH;
                    ddbb+=dispwidth;
                }
            }


            ddrval = IDirectDrawSurface_Unlock(lpBackBuffer,NULL);
            while(ddrval == DDERR_WASSTILLDRAWING||ddrval==DDERR_SURFACEBUSY)//so it doesn't die if it is too quick.
                ddrval = IDirectDrawSurface_Unlock(lpBackBuffer,NULL);

            ddrval = IDirectDrawSurface_Flip(lpFrontBuffer, NULL, DDFLIP_WAIT);
            while(ddrval == DDERR_WASSTILLDRAWING||ddrval==DDERR_SURFACEBUSY)//again,so it doesn't die if it is too quick.
                ddrval = IDirectDrawSurface_Flip(lpFrontBuffer, NULL, DDFLIP_WAIT);

            if( ddrval != DD_OK )
            {
                ReportError("Flip failed ddrval = ", ddrval );
                return ;
            }
        }
    }
    //Print("End Frame Swap");

    return;

} /* DDClear */

/*
 * DDCreateFlippingSurface
 *
 * create a FrontBuffer and a BackBuffer(s)
 *
 */
BOOL DDCreateFlippingSurface( void )
{
//    DDPIXELFORMAT       ddpf;
    DDSURFACEDESC	ddsd;
    HRESULT		ddrval;
    DDSCAPS		ddscaps;
    DDCAPS              ddcaps;
    char mys[1024];
    ddcaps.dwSize = sizeof( ddcaps );

    if( IDirectDraw_GetCaps( lpDD, &ddcaps, NULL ) != DD_OK )
        return FALSE;

    /*
     * fill in surface desc:
     * want a primary surface with 2 back buffers
     */
    memset( &ddsd, 0,sizeof( ddsd ) );
    ddsd.dwSize = sizeof( ddsd );

    if (nBufferCount > 1)
    {
        //
        //  fullscreen case, create a primary (ie front) and
        //  either 1  or 2 back buffers
        //
        ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
        ddsd.dwBackBufferCount = nBufferCount-1;
        ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                DDSCAPS_FLIP | DDSCAPS_COMPLEX;

        //OutputDebugString("Creating multiple backbuffer primary\n\r");
        ddrval = IDirectDraw_CreateSurface( lpDD, &ddsd, &lpFrontBuffer, NULL );

        if( ddrval != DD_OK )
        {
            sprintf(mys,"%i %i %i\n%i %i %i\n%i %i %i\n%i %i %i\n%i %i %i\n%i %i %i\n",
DDERR_INCOMPATIBLEPRIMARY,DDERR_INVALIDCAPS,DDERR_INVALIDOBJECT ,
DDERR_INVALIDPARAMS,DDERR_INVALIDPIXELFORMAT,DDERR_NOALPHAHW ,
DDERR_NOCOOPERATIVELEVELSET,DDERR_NODIRECTDRAWHW,DDERR_NOEMULATION ,
DDERR_NOEXCLUSIVEMODE,DDERR_NOFLIPHW,DDERR_NOMIPMAPHW,
DDERR_NOOVERLAYHW ,DDERR_NOZBUFFERHW ,DDERR_OUTOFMEMORY ,
DDERR_OUTOFVIDEOMEMORY ,DDERR_PRIMARYSURFACEALREADYEXISTS ,DDERR_UNSUPPORTEDMODE);
            Print(mys);
            ReportError( "CreateSurface FAILED! ", ddrval );
            return FALSE;
        }

        /*
         * go find the back buffer
         */
        ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
        ddrval = IDirectDrawSurface_GetAttachedSurface(
                    lpFrontBuffer,
                    &ddscaps,
                    &lpBackBuffer );

        if( ddrval != DD_OK )
        {
            ReportError( "GetAttachedSurface failed! ",ddrval );
            return FALSE;
        }

    }
    IDirectDrawSurface_GetSurfaceDesc(lpFrontBuffer,&ddsd);
    dispwidth=ddsd.lPitch;

    DDClear();

} /* DDCreateFlippingSurface */

/*
 * DDCreateSurface
 */
LPDIRECTDRAWSURFACE DDCreateSurface(
		DWORD width,
		DWORD height,
		BOOL sysmem,
		BOOL trans )
{
    DDSURFACEDESC	ddsd;
    HRESULT		ddrval;
//    DDCOLORKEY		ddck;
    LPDIRECTDRAWSURFACE	psurf;


    /*
     * fill in surface desc
     */
    memset( &ddsd, 0, sizeof( ddsd ) );
    ddsd.dwSize = sizeof( ddsd );
    ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT |DDSD_WIDTH;

    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    if( sysmem || bUseSysMem )
    {
	ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
    }

    ddsd.dwHeight = height;
    ddsd.dwWidth = width;

    ddrval = IDirectDraw_CreateSurface( lpDD, &ddsd, &psurf, NULL );

    /*
     * set the color key for this bitmap
     */
    if( ddrval == DD_OK )
    {}
    else
    {
	ReportError( "CreateSurface FAILED, rc ", (DWORD) LOWORD( ddrval ) );
	psurf = NULL;
    }

     return psurf;

} /* DDCreateSurface */

/*
 * ReadPalFile
 *
 * Create a DirectDrawPalette from a palette file
 *
 * if the palette files cant be found, make a default 332 palette
 */
char mys[64];
void DD_LoadPal( char *pal)
{
    int                 i;
//    int			fh;
    HRESULT		ddrval;


    //Print("Palette");
    for( i=0; i<256; i++ )
    {
        ape[i].peRed   = (BYTE)(pal[i*3+0]);
        ape[i].peGreen = (BYTE)(pal[i*3+1]);
        ape[i].peBlue  = (BYTE)(pal[i*3+2]);
        ape[i].peFlags = (BYTE)0;
    }
    //Print("Hi");
    if(!ppal){
        ddrval = IDirectDraw_CreatePalette(
			    lpDD,
			    DDPCAPS_8BIT,
                ape,
			    &ppal,
			    NULL );
        //sprintf(mys,"Create: %i",ddrval);
        //Print(mys);
        ddrval = IDirectDrawSurface_SetPalette(lpFrontBuffer,ppal);
        ddrval = IDirectDrawSurface_SetPalette(lpBackBuffer,ppal);

/*        sprintf(mys,"%i",ddrval);
        Print(mys);*/
/*        sprintf(mys,"%i",ddrval);
        Print(mys);*/
    }
    else{
        ddrval = IDirectDrawPalette_SetEntries(ppal,0, 0, 256, ape);
        //sprintf(mys,"Update: %i",ddrval);
        //Print(mys);
        /*sprintf(mys,"Errors: %i %i %i %i %i",
        DDERR_INVALIDOBJECT,
        DDERR_INVALIDPARAMS,
        DDERR_NOPALETTEATTACHED,
        DDERR_NOTPALETTIZED,
        DDERR_UNSUPPORTED);
        Print(mys);*/
    }
    //Print("Bye");
/*    sprintf(mys,"Errors: %i %i %i %i %i %i %i",
    DDERR_INVALIDOBJECT,
    DDERR_INVALIDPARAMS,
    DDERR_NOCOOPERATIVELEVELSET,
    DDERR_NOEXCLUSIVEMODE,
    DDERR_OUTOFCAPS,
    DDERR_OUTOFMEMORY,
    DDERR_UNSUPPORTED);*/
    //Print(mys);

    //return ppal;

} /* ReadPalFile */


BOOL DDClear( void )
{
    DDBLTFX     ddbltfx;
    int         i;
    HRESULT     ddrval;

    if (bFullscreen)
    {
    UpdateWindow(mainhwnd);

    ddbltfx.dwSize = sizeof( ddbltfx );
    ddbltfx.dwFillColor = 0;
        /*
         * do it for all buffers, we either have 1 or 2 back buffers
         * make sure we get them all, 4 is plenty!
         */
        for( i=0; i<4; i++ )
        {
            ddrval = IDirectDrawSurface_Blt(
                            lpBackBuffer,           // dest surface
                            NULL,                   // dest rect
                            NULL,                   // src surface
                            NULL,                   // src rect
                            DDBLT_COLORFILL | DDBLT_WAIT,
                            &ddbltfx);

            if( ddrval != DD_OK )
            {
                Print("Fill failed ddrval");
                return FALSE;
            }

            ddrval = IDirectDrawSurface_Flip(lpFrontBuffer, NULL, DDFLIP_WAIT);

            if( ddrval != DD_OK )
            {
                Print("Flip failed ddrval" );
                return FALSE;
            }
        }
    }

    return TRUE;

} /* DDClear */
